# SPDX-FileCopyrightText: 2023 Julian Schauder <pytraeger@schauder.info>
#
# SPDX-License-Identifier: MIT

import threading
import time
import logging
import os

from pytraeger.mqtt import Mqtt
from pytraeger.api import Api


_LOGGER: logging.Logger = logging.getLogger(__package__)
_LOGGER.setLevel(logging.DEBUG)


class Manager:
    def __init__(self,
                 username="",
                 password="",
                 interval_idle=60*5,
                 interval_busy=10,
                 ):
        if username == "":
            self.username = os.environ["PYTRAEGER_USERNAME"]
        else:
            self.username = username
        if password == "":
            self.password = os.environ["PYTRAEGER_PASSWORD"]
        else:
            self.password = password

        self.interval_idle = interval_idle
        self.interval_busy = interval_busy
        self.api = Api(self.username, self.password,
                       self.interval_idle, self.interval_busy)
        self.mqtt_client = Mqtt(self.api.get_mqtt_endpoint(),
                                [grill.identifier for grill in self.api.grills])
        self.mqtt_client.set_datahook(self.mqtt_event_dispatcher)

        self.start_thread_apipolling()

    def start_thread_apipolling(self):
        for grill in self.api.grills:
            threading.Thread(target=poll_updates, args=(
                grill.identifier, self.api)).start()
        threading.Thread(target=watchdog, args=()).start()

    def mqtt_event_dispatcher(self, topic, data):
        for grill in self.api.grills:
            if topic.endswith(grill.identifier):
                grill.push_data(data)
                _LOGGER.debug("Message Received from Grill(%s): Online : %s",
                              topic, data["status"]["connected"]
                              )
                break


def watchdog():
    # After two hours, kill the program.
    # The current implementation should die after exactly 1h due to outdated tokens.
    # If an unexpected bug prevents natural death by age, initiate death by watchdog
    time.sleep(60*60*2)
    os._exit(1)


def poll_updates(grill_identifier, parent_api: Api):
    # Make sure to pull the grill via the identifier.
    # Might only be usefull for grill reinstanciations.
    while True:
        try:
            grill_object = None
            for grill in parent_api.grills:
                if grill.identifier == grill_identifier:
                    grill_object = grill
                    break

            if grill_object.is_on:
                time.sleep(parent_api.interval_busy)
            else:
                time.sleep(parent_api.interval_idle)
            _LOGGER.debug("Polling Update for %s",
                          grill_identifier
                          )

            parent_api.update_grill(grill_identifier)
        except Exception as exception:
            _LOGGER.critical(exception)
            os._exit(1)



# SPDX-FileCopyrightText: 2023 Julian Schauder <pytraeger@schauder.info>
#
# SPDX-License-Identifier: MIT

import urllib

import logging
import ssl
import json

import paho.mqtt.client as _mqtt

_LOGGER: logging.Logger = logging.getLogger(__package__)
_LOGGER.setLevel(logging.DEBUG)


class Mqtt:
    def __init__(self, connectstring, grills, immediate_init=True):
        self.connectstring = connectstring
        self.grills = grills
        self.client = None
        self.datahook = None
        if immediate_init:
            self.init()
            self.start()

    def start(self):
        self.client.loop_start()

    def set_datahook(self, hook):
        self.datahook = hook

    def mqtt_onconnect(self, client_instance, userdata, flags, rc):
        _LOGGER.debug("Client Connected")

        for grill in self.grills:
            client_instance.subscribe(
                ("prod/thing/update/{}".format(grill), 1))
        _LOGGER.debug("Grill Connected")

    def mqtt_onconnectfail(self, client, userdata):
        _LOGGER.debug(
            "Connect Fail Callback. Client:%s userdata: %s", client, userdata)
        _LOGGER.debug("Grill Connect Failed! MQTT Client Kill.")

    def mqtt_onlog(self, client, userdata, level, buf):
        _LOGGER.debug(
            "OnLog Callback. Client:%s userdata:%s level:%s buf: %s", 
            client, userdata, level, buf)

    def mqtt_onsubscribe(self, client, userdata, mid, granted_qos):
        _LOGGER.debug(
            "OnSubscribe Callback. Client:%s userdata:%s mid:%s granted_qos:%s", 
            client, userdata, mid, granted_qos)

    def mqtt_onmessage(self, client, userdata, message):
        _LOGGER.debug("Message for Grill: %s", message.topic)

        data = json.loads(message.payload)
        try:
            self.datahook(message.topic, data)
        except Exception as e:
            _LOGGER.warning(e)

    def init(self):
        mqtt_client = _mqtt.Client(
            transport="websockets", reconnect_on_failure=False, clean_session=True)
        mqtt_client.on_connect = self.mqtt_onconnect
        mqtt_client.on_connect_fail = self.mqtt_onconnectfail
        mqtt_client.enable_logger(_LOGGER)
        mqtt_client.on_subscribe = self.mqtt_onsubscribe
        mqtt_client.on_message = self.mqtt_onmessage
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        mqtt_client.tls_set_context(context)
        mqtt_client.reconnect_delay_set(min_delay=3, max_delay=5)
        mqtt_parts = urllib.parse.urlparse(self.connectstring)

        headers = {
            "Host": "{0:s}".format(mqtt_parts.netloc),
        }
        mqtt_client.ws_set_options(path="{}?{}".format(
            mqtt_parts.path, mqtt_parts.query), headers=headers)
        _LOGGER.debug("Connecting to %s", mqtt_parts.netloc)
        mqtt_client.connect(mqtt_parts.netloc, port=443, keepalive=30)
        self.client = mqtt_client



# SPDX-FileCopyrightText: 2023 Julian Schauder <pytraeger@schauder.info>
#
# SPDX-License-Identifier: MIT

import datetime
import logging

import requests

from pytraeger.grill import grill as grill_class
from pytraeger.mqtt import Mqtt

AMAZON_COGNITO_GATEWAY = 'https://cognito-idp.us-west-2.amazonaws.com/'
AMAZON_API_ENDPOINT = 'https://1ywgyc65d1.execute-api.us-west-2.amazonaws.com/prod/'
TRAEGER_CLIENT_ID = "2fuohjtqv1e63dckp5v84rau0j"  # Seems to be the Traeger-ID

AMAZON_COMMAND_CODES = {
    "refresh": 90
}
HTTP_TIMEOUT = 20
_LOGGER: logging.Logger = logging.getLogger(__package__)
_LOGGER.setLevel(logging.DEBUG)


class Api:
    def __init__(self,
                 username="",
                 password="",
                 interval_idle=60*5,
                 interval_busy=10,
                 immediate_init=True
                 ):
        self.username = username
        self.password = password

        self.interval_idle = interval_idle
        self.interval_busy = interval_busy
        self.grills = [grill_class]  # will be of type pytraeger.grill

        self.cognito_accesstoken = None

        self.mqtt_endpoint = None
        self.mqtt_client = None
        if immediate_init:
            self.init()

    def init(self):
        _LOGGER.debug("Initialization of API")
        self.do_cognito()
        self.get_grills()

        # Asking Grills for an Update is nonessential but ensures a quick application boot
        for grill in self.grills:
            self.update_grill(grill.identifier)

    def get_amzdate(self):
        return datetime.datetime.utcnow().strftime('%Y%m%dT%H%M%SZ')

    def do_cognito(self):
        _LOGGER.debug("Cognito Initiation")
        data = {
            "ClientMetadata": {},
            "AuthParameters": {
                "PASSWORD": self.password,
                "USERNAME": self.username,
            },
            "AuthFlow": "USER_PASSWORD_AUTH",
            "ClientId": TRAEGER_CLIENT_ID
        }
        response = requests.post(AMAZON_COGNITO_GATEWAY,
                                 json=data,
                                 headers={'Content-Type': 'application/x-amz-json-1.1',
                                          'X-Amz-Date': self.get_amzdate(),
                                          'X-Amz-Target': 'AWSCognitoIdentityProviderService.InitiateAuth'
                                          }, timeout=HTTP_TIMEOUT)
        assert response.status_code == 200

        self.cognito_accesstoken = response.json(
        )["AuthenticationResult"]["AccessToken"]
        _LOGGER.debug("Login successful")
        # TODO- Validity

    def get_grills(self):
        _LOGGER.debug("Inventory-Update")
        response = requests.get(AMAZON_API_ENDPOINT+"/users/self",
                                headers={
                                    'authorization': self.cognito_accesstoken}, timeout=HTTP_TIMEOUT)
        assert response.status_code == 200
        grills = response.json()["things"]
        self.grills = []
        for grill in grills:
            self.grills.append(grill_class(grill["thingName"]))
            _LOGGER.debug("Added ID: %s",
                          grill["thingName"]
                          )

    def update_grill(self, grill_identifier):
        response = requests.post(AMAZON_API_ENDPOINT+"/things/{}/commands".format(grill_identifier),
                                 json={
            'command': AMAZON_COMMAND_CODES["refresh"]
        },
            headers={
            'Authorization': self.cognito_accesstoken,
            "Content-Type": "application/json",
            "Accept-Language": "en-us",
            'X-Amz-Date': self.get_amzdate(),
            "User-Agent": "Traeger/11 CFNetwork/1209 Darwin/20.2.0",
        }, timeout=HTTP_TIMEOUT)
        assert response.status_code == 200

    def get_mqtt_endpoint(self):
        self.mqtt_endpoint = requests.post(AMAZON_API_ENDPOINT+"/mqtt-connections",
                                           headers={
                                               'Authorization': self.cognito_accesstoken,
                                               'X-Amz-Date': self.get_amzdate(),
                                               "Content-Type": "application/json",
                                               "Accept-Language": "en-us",
                                               "User-Agent": "Traeger/11 CFNetwork/1209 Darwin/20.2.0",
                                           }, timeout=HTTP_TIMEOUT).json()["signedUrl"]
        return self.mqtt_endpoint



# SPDX-FileCopyrightText: 2023 Julian Schauder <pytraeger@schauder.info>
#
# SPDX-License-Identifier: MIT

class grill:
    def __init__(self, identifier):
        self.identifier = identifier
        self.data = {}
        self.last_update = None
        self.is_on = True # Default Value of "ON" might speed-up initalization phase as it is primarily used for pollscheduling
        self.listeners = []
    
    def push_data(self, data):
        self.data = data
        self.is_on = data["status"]["connected"]
        for listener in self.listeners:
            try:
                listener(self)
            except:
                pass

    def register_listener(self, listener):
        self.listeners.append(listener)


    def __str__(self):
        return self.identifier